@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# Copyright 2013 Matt Martz
+# Copyright 2012-2014 Matt Martz
 # All Rights Reserved.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
@@ -15,21 +15,44 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-__version__ = '0.2.2'
+import os
+import re
+import sys
+import math
+import signal
+import socket
+import timeit
+import threading
+
+__version__ = '0.3.2'
+
+# Some global variables we use
+user_agent = 'speedtest-cli/%s' % __version__
+source = None
+shutdown_event = None
+
+# Used for bound_interface
+socket_socket = socket.socket
 
 try:
-    from urllib2 import urlopen, Request
+    import xml.etree.cElementTree as ET
 except ImportError:
-    from urllib.request import urlopen, Request
+    try:
+        import xml.etree.ElementTree as ET
+    except ImportError:
+        from xml.dom import minidom as DOM
+        ET = None
 
-import math
-import time
-import os
-import sys
-import threading
-import re
-import signal
-from xml.dom import minidom as DOM
+# Begin import game to handle Python 2 and Python 3
+try:
+    from urllib2 import urlopen, Request, HTTPError, URLError
+except ImportError:
+    from urllib.request import urlopen, Request, HTTPError, URLError
+
+try:
+    from httplib import HTTPConnection, HTTPSConnection
+except ImportError:
+    from http.client import HTTPConnection, HTTPSConnection
 
 try:
     from Queue import Queue
@@ -116,6 +139,22 @@ else:
     del builtins
 
 
+class SpeedtestCliServerListError(Exception):
+    """Internal Exception class used to indicate to move on to the next
+    URL for retrieving speedtest.net server details
+
+    """
+
+
+def bound_socket(*args, **kwargs):
+    """Bind socket to a specified source IP address"""
+
+    global source
+    sock = socket_socket(*args, **kwargs)
+    sock.bind((source, 0))
+    return sock
+
+
 def distance(origin, destination):
     """Determine distance between 2 sets of [lat,lon] in km"""
 
@@ -125,16 +164,43 @@ def distance(origin, destination):
 
     dlat = math.radians(lat2 - lat1)
     dlon = math.radians(lon2 - lon1)
-    a = (math.sin(dlat / 2) * math.sin(dlat / 2) + math.cos(math.radians(lat1))
-         * math.cos(math.radians(lat2)) * math.sin(dlon / 2)
-         * math.sin(dlon / 2))
+    a = (math.sin(dlat / 2) * math.sin(dlat / 2) +
+         math.cos(math.radians(lat1)) *
+         math.cos(math.radians(lat2)) * math.sin(dlon / 2) *
+         math.sin(dlon / 2))
     c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
     d = radius * c
 
     return d
 
 
+def build_request(url, data=None, headers={}):
+    """Build a urllib2 request object
+
+    This function automatically adds a User-Agent header to all requests
+
+    """
+
+    headers['User-Agent'] = user_agent
+    return Request(url, data=data, headers=headers)
+
+
+def catch_request(request):
+    """Helper function to catch common exceptions encountered when
+    establishing a connection with a HTTP/HTTPS request
+
+    """
+
+    try:
+        uh = urlopen(request)
+        return uh
+    except (HTTPError, URLError, socket.error):
+        return False
+
+
 class FileGetter(threading.Thread):
+    """Thread class for retrieving a URL"""
+
     def __init__(self, url, start):
         self.url = url
         self.result = None
@@ -144,8 +210,9 @@ class FileGetter(threading.Thread):
     def run(self):
         self.result = [0]
         try:
-            if (time.time() - self.starttime) <= 10:
-                f = urlopen(self.url)
+            if (timeit.default_timer() - self.starttime) <= 10:
+                request = build_request(self.url)
+                f = urlopen(request)
                 while 1 and not shutdown_event.isSet():
                     self.result.append(len(f.read(10240)))
                     if self.result[-1] == 0:
@@ -156,7 +223,9 @@ class FileGetter(threading.Thread):
 
 
 def downloadSpeed(files, quiet=False):
-    start = time.time()
+    """Function to launch FileGetter threads and calculate download speeds"""
+
+    start = timeit.default_timer()
 
     def producer(q, files):
         for file in files:
@@ -180,17 +249,19 @@ def downloadSpeed(files, quiet=False):
     q = Queue(6)
     prod_thread = threading.Thread(target=producer, args=(q, files))
     cons_thread = threading.Thread(target=consumer, args=(q, len(files)))
-    start = time.time()
+    start = timeit.default_timer()
     prod_thread.start()
     cons_thread.start()
     while prod_thread.isAlive():
         prod_thread.join(timeout=0.1)
     while cons_thread.isAlive():
         cons_thread.join(timeout=0.1)
-    return (sum(finished) / (time.time() - start))
+    return (sum(finished) / (timeit.default_timer() - start))
 
 
 class FilePutter(threading.Thread):
+    """Thread class for putting a URL"""
+
     def __init__(self, url, start, size):
         self.url = url
         chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
@@ -203,9 +274,10 @@ class FilePutter(threading.Thread):
 
     def run(self):
         try:
-            if ((time.time() - self.starttime) <= 10 and
+            if ((timeit.default_timer() - self.starttime) <= 10 and
                     not shutdown_event.isSet()):
-                f = urlopen(self.url, self.data)
+                request = build_request(self.url, data=self.data)
+                f = urlopen(request)
                 f.read(11)
                 f.close()
                 self.result = len(self.data)
@@ -216,7 +288,9 @@ class FilePutter(threading.Thread):
 
 
 def uploadSpeed(url, sizes, quiet=False):
-    start = time.time()
+    """Function to launch FilePutter threads and calculate upload speeds"""
+
+    start = timeit.default_timer()
 
     def producer(q, sizes):
         for size in sizes:
@@ -240,17 +314,23 @@ def uploadSpeed(url, sizes, quiet=False):
     q = Queue(6)
     prod_thread = threading.Thread(target=producer, args=(q, sizes))
     cons_thread = threading.Thread(target=consumer, args=(q, len(sizes)))
-    start = time.time()
+    start = timeit.default_timer()
     prod_thread.start()
     cons_thread.start()
     while prod_thread.isAlive():
         prod_thread.join(timeout=0.1)
     while cons_thread.isAlive():
         cons_thread.join(timeout=0.1)
-    return (sum(finished) / (time.time() - start))
+    return (sum(finished) / (timeit.default_timer() - start))
 
 
 def getAttributesByTagName(dom, tagName):
+    """Retrieve an attribute from an XML document and return it in a
+    consistent format
+
+    Only used with xml.dom.minidom, which is likely only to be used
+    with python versions older than 2.5
+    """
     elem = dom.getElementsByTagName(tagName)[0]
     return dict(list(elem.attributes.items()))
 
@@ -260,19 +340,39 @@ def getConfig():
     we are interested in
     """
 
-    uh = urlopen('http://www.speedtest.net/speedtest-config.php')
-    configxml = uh.read()
+    request = build_request('https://www.speedtest.net/speedtest-config.php')
+    uh = catch_request(request)
+    if uh is False:
+        print_('Could not retrieve speedtest.net configuration')
+        sys.exit(1)
+    configxml = []
+    while 1:
+        configxml.append(uh.read(10240))
+        if len(configxml[-1]) == 0:
+            break
     if int(uh.code) != 200:
         return None
     uh.close()
-    root = DOM.parseString(configxml)
-    config = {
-        'client': getAttributesByTagName(root, 'client'),
-        'times': getAttributesByTagName(root, 'times'),
-        'download': getAttributesByTagName(root, 'download'),
-        'upload': getAttributesByTagName(root, 'upload')}
-
+    try:
+        try:
+            root = ET.fromstring(''.encode().join(configxml))
+            config = {
+                'client': root.find('client').attrib,
+                'times': root.find('times').attrib,
+                'download': root.find('download').attrib,
+                'upload': root.find('upload').attrib}
+        except AttributeError:  # Python3 branch
+            root = DOM.parseString(''.join(configxml))
+            config = {
+                'client': getAttributesByTagName(root, 'client'),
+                'times': getAttributesByTagName(root, 'times'),
+                'download': getAttributesByTagName(root, 'download'),
+                'upload': getAttributesByTagName(root, 'upload')}
+    except SyntaxError:
+        print_('Failed to parse speedtest.net configuration')
+        sys.exit(1)
     del root
+    del configxml
     return config
 
 
@@ -281,23 +381,62 @@ def closestServers(client, all=False):
     distance
     """
 
-    uh = urlopen('http://www.speedtest.net/speedtest-servers.php')
-    serversxml = uh.read()
-    if int(uh.code) != 200:
-        return None
-    uh.close()
-    root = DOM.parseString(serversxml)
+    urls = [
+        'https://www.speedtest.net/speedtest-servers-static.php',
+        'http://c.speedtest.net/speedtest-servers-static.php',
+    ]
     servers = {}
-    for server in root.getElementsByTagName('server'):
-        attrib = dict(list(server.attributes.items()))
-        d = distance([float(client['lat']), float(client['lon'])],
-                     [float(attrib.get('lat')), float(attrib.get('lon'))])
-        attrib['d'] = d
-        if d not in servers:
-            servers[d] = [attrib]
-        else:
-            servers[d].append(attrib)
-    del root
+    for url in urls:
+        try:
+            request = build_request(url)
+            uh = catch_request(request)
+            if uh is False:
+                raise SpeedtestCliServerListError
+            serversxml = []
+            while 1:
+                serversxml.append(uh.read(10240))
+                if len(serversxml[-1]) == 0:
+                    break
+            if int(uh.code) != 200:
+                uh.close()
+                raise SpeedtestCliServerListError
+            uh.close()
+            try:
+                try:
+                    root = ET.fromstring(''.encode().join(serversxml))
+                    elements = root.getiterator('server')
+                except AttributeError:  # Python3 branch
+                    root = DOM.parseString(''.join(serversxml))
+                    elements = root.getElementsByTagName('server')
+            except SyntaxError:
+                raise SpeedtestCliServerListError
+            for server in elements:
+                try:
+                    attrib = server.attrib
+                except AttributeError:
+                    attrib = dict(list(server.attributes.items()))
+                d = distance([float(client['lat']),
+                              float(client['lon'])],
+                             [float(attrib.get('lat')),
+                              float(attrib.get('lon'))])
+                attrib['d'] = d
+                if d not in servers:
+                    servers[d] = [attrib]
+                else:
+                    servers[d].append(attrib)
+            del root
+            del serversxml
+            del elements
+        except SpeedtestCliServerListError:
+            continue
+
+        # We were able to fetch and parse the list of speedtest.net servers
+        if servers:
+            break
+
+    if not servers:
+        print_('Failed to retrieve list of speedtest.net servers')
+        sys.exit(1)
 
     closest = []
     for d in sorted(servers.keys()):
@@ -314,27 +453,37 @@ def closestServers(client, all=False):
 
 
 def getBestServer(servers):
-    """Perform a speedtest.net "ping" to determine which speedtest.net
-    server has the lowest latency
+    """Perform a speedtest.net latency request to determine which
+    speedtest.net server has the lowest latency
     """
 
     results = {}
     for server in servers:
         cum = []
-        url = os.path.dirname(server['url'])
+        url = '%s/latency.txt' % os.path.dirname(server['url'])
+        urlparts = urlparse(url)
         for i in range(0, 3):
-            uh = urlopen('%s/latency.txt' % url)
-            start = time.time()
-            text = uh.read(9)
-            total = time.time() - start
-            if int(uh.code) == 200 and text == 'test=test'.encode():
+            try:
+                if urlparts[0] == 'https':
+                    h = HTTPSConnection(urlparts[1])
+                else:
+                    h = HTTPConnection(urlparts[1])
+                headers = {'User-Agent': user_agent}
+                start = timeit.default_timer()
+                h.request("GET", urlparts[2], headers=headers)
+                r = h.getresponse()
+                total = (timeit.default_timer() - start)
+            except (HTTPError, URLError, socket.error):
+                cum.append(3600)
+                continue
+            text = r.read(9)
+            if int(r.status) == 200 and text == 'test=test'.encode():
                 cum.append(total)
             else:
                 cum.append(3600)
-            uh.close()
-        avg = round((sum(cum) / 3) * 1000000, 3)
+            h.close()
+        avg = round((sum(cum) / 6) * 1000, 3)
         results[avg] = server
-
     fastest = sorted(results.keys())[0]
     best = results[fastest]
     best['latency'] = fastest
@@ -343,19 +492,25 @@ def getBestServer(servers):
 
 
 def ctrl_c(signum, frame):
+    """Catch Ctrl-C key sequence and set a shutdown_event for our threaded
+    operations
+    """
+
     global shutdown_event
     shutdown_event.set()
     raise SystemExit('\nCancelling...')
 
 
 def version():
+    """Print the version"""
+
     raise SystemExit(__version__)
 
 
 def speedtest():
     """Run the full speedtest.net test"""
 
-    global shutdown_event
+    global shutdown_event, source
     shutdown_event = threading.Event()
 
     signal.signal(signal.SIGINT, ctrl_c)
@@ -368,10 +523,16 @@ def speedtest():
         'https://github.com/sivel/speedtest-cli')
 
     parser = ArgParser(description=description)
+    # Give optparse.OptionParser an `add_argument` method for
+    # compatibility with argparse.ArgumentParser
     try:
         parser.add_argument = parser.add_option
     except AttributeError:
         pass
+    parser.add_argument('--bytes', dest='units', action='store_const',
+                        const=('byte', 1), default=('bit', 8),
+                        help='Display values in bytes instead of bits. Does '
+                             'not affect the image generated by --share')
     parser.add_argument('--share', action='store_true',
                         help='Generate and provide a URL to the speedtest.net '
                              'share results image')
@@ -383,6 +544,9 @@ def speedtest():
                              'sorted by distance')
     parser.add_argument('--server', help='Specify a server ID to test against')
     parser.add_argument('--mini', help='URL of the Speedtest Mini server')
+    parser.add_argument('--source', help='Source IP address to bind to')
+    parser.add_argument('--timeout', default=10, type=int,
+                        help='HTTP timeout in seconds. Default 10')
     parser.add_argument('--version', action='store_true',
                         help='Show the version number and exit')
 
@@ -393,12 +557,24 @@ def speedtest():
         args = options
     del options
 
+    # Print the version and exit
     if args.version:
         version()
 
+    socket.setdefaulttimeout(args.timeout)
+
+    # If specified bind to a specific IP address
+    if args.source:
+        source = args.source
+        socket.socket = bound_socket
+
     if not args.simple:
         print_('Retrieving speedtest.net configuration...')
-    config = getConfig()
+    try:
+        config = getConfig()
+    except URLError:
+        print_('Cannot retrieve speedtest configuration')
+        sys.exit(1)
 
     if not args.simple:
         print_('Retrieving speedtest.net server list...')
@@ -410,8 +586,14 @@ def speedtest():
                 line = ('%(id)4s) %(sponsor)s (%(name)s, %(country)s) '
                         '[%(d)0.2f km]' % server)
                 serverList.append(line)
+            # Python 2.7 and newer seem to be ok with the resultant encoding
+            # from parsing the XML, but older versions have some issues.
+            # This block should detect whether we need to encode or not
             try:
+                unicode()
                 print_('\n'.join(serverList).encode('utf-8', 'ignore'))
+            except NameError:
+                print_('\n'.join(serverList))
             except IOError:
                 pass
             sys.exit(0)
@@ -436,7 +618,8 @@ def speedtest():
             url = args.mini
         urlparts = urlparse(url)
         try:
-            f = urlopen(args.mini)
+            request = build_request(args.mini)
+            f = urlopen(request)
         except:
             print_('Invalid Speedtest Mini URL')
             sys.exit(1)
@@ -444,6 +627,21 @@ def speedtest():
             text = f.read()
             f.close()
         extension = re.findall('upload_extension: "([^"]+)"', text.decode())
+        if not extension:
+            for ext in ['php', 'asp', 'aspx', 'jsp']:
+                try:
+                    request = build_request('%s/speedtest/upload.%s' %
+                                            (args.mini, ext))
+                    f = urlopen(request)
+                except:
+                    pass
+                else:
+                    data = f.read().strip()
+                    if (f.code == 200 and
+                            len(data.splitlines()) == 1 and
+                            re.match('size=[0-9]', data)):
+                        extension = [ext]
+                        break
         if not urlparts or not extension:
             print_('Please provide the full URL of your Speedtest Mini server')
             sys.exit(1)
@@ -461,12 +659,20 @@ def speedtest():
             best = servers[0]
     else:
         if not args.simple:
-            print_('Selecting best server based on ping...')
+            print_('Selecting best server based on latency...')
         best = getBestServer(servers)
 
     if not args.simple:
-        print_('Hosted by %(sponsor)s (%(name)s) [%(d)0.2f km]: '
-               '%(latency)s ms' % best)
+        # Python 2.7 and newer seem to be ok with the resultant encoding
+        # from parsing the XML, but older versions have some issues.
+        # This block should detect whether we need to encode or not
+        try:
+            unicode()
+            print_(('Hosted by %(sponsor)s (%(name)s) [%(d)0.2f km]: '
+                   '%(latency)s ms' % best).encode('utf-8', 'ignore'))
+        except NameError:
+            print_('Hosted by %(sponsor)s (%(name)s) [%(d)0.2f km]: '
+                   '%(latency)s ms' % best)
     else:
         print_('Ping: %(latency)s ms' % best)
 
@@ -481,7 +687,8 @@ def speedtest():
     dlspeed = downloadSpeed(urls, args.simple)
     if not args.simple:
         print_()
-    print_('Download: %0.2f Mbit/s' % ((dlspeed / 1000 / 1000) * 8))
+    print_('Download: %0.2f M%s/s' %
+           ((dlspeed / 1000 / 1000) * args.units[1], args.units[0]))
 
     sizesizes = [int(.25 * 1000 * 1000), int(.5 * 1000 * 1000)]
     sizes = []
@@ -493,7 +700,8 @@ def speedtest():
     ulspeed = uploadSpeed(best['url'], sizes, args.simple)
     if not args.simple:
         print_()
-    print_('Upload: %0.2f Mbit/s' % ((ulspeed / 1000 / 1000) * 8))
+    print_('Upload: %0.2f M%s/s' %
+           ((ulspeed / 1000 / 1000) * args.units[1], args.units[0]))
 
     if args.share and args.mini:
         print_('Cannot generate a speedtest.net share results image while '
@@ -503,6 +711,9 @@ def speedtest():
         ping = int(round(best['latency'], 0))
         ulspeedk = int(round((ulspeed / 1000) * 8, 0))
 
+        # Build the request to send results back to speedtest.net
+        # We use a list instead of a dict because the API expects parameters
+        # in a certain order
         apiData = [
             'download=%s' % dlspeedk,
             'ping=%s' % ping,
@@ -516,10 +727,14 @@ def speedtest():
                              (ping, ulspeedk, dlspeedk, '297aae72'))
                             .encode()).hexdigest()]
 
-        req = Request('http://www.speedtest.net/api/api.php',
-                      data='&'.join(apiData).encode())
-        req.add_header('Referer', 'http://c.speedtest.net/flash/speedtest.swf')
-        f = urlopen(req)
+        headers = {'Referer': 'https://c.speedtest.net/flash/speedtest.swf'}
+        request = build_request('https://www.speedtest.net/api/api.php',
+                                data='&'.join(apiData).encode(),
+                                headers=headers)
+        f = catch_request(request)
+        if f is False:
+            print_('Could not submit results to speedtest.net')
+            sys.exit(1)
         response = f.read()
         code = f.code
         f.close()
@@ -534,7 +749,7 @@ def speedtest():
             print_('Could not submit results to speedtest.net')
             sys.exit(1)
 
-        print_('Share results: http://www.speedtest.net/result/%s.png' %
+        print_('Share results: https://www.speedtest.net/result/%s.png' %
                resultid[0])
 
 
